/* $Id: bdiffm.c 1720 2005-07-28 19:34:58Z $
 * bdiffm - display binary diff matrix for n files.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define uint unsigned int
#define byte unsigned char

#define min(a, b) (a < b ? a : b)
#define max(a, b) (a > b ? a : b)
#define abs(a) a >= 0 ? a : -(a)

#define CL_RESET 0
#define CL_RED 31
#define CL_GREEN 32
#define CL_YELLOW 33
#define CL_WHITE 37

void
pcolor (uint c)
{
  if (c == CL_RESET)
    printf ("\033[0m");
  else
    printf ("\033[%d;1m", c);
}

uint
getFileSize (const char *filename)
{
  uint size = 0;
  FILE *f = fopen (filename, "r");

  if (f)
    {
      fseek (f, 0, SEEK_END);
      size = ftell (f);
      fclose (f);
    }

  return size;
}

uint
getSimilarity (const char *a, const char *b)
{
  char *command;
  FILE *f;
  char buffer[0x10];
  uint bytesRead;
  uint sizeA, sizeB;
  uint diffBytes;

  sizeA = getFileSize (a);
  sizeB = getFileSize (b);

  if (max (sizeA, sizeB) == 0)
    return 0;

  asprintf (&command, "cmp -bl %s %s 2> /dev/null | wc -l ", a, b);
  f = popen (command, "r");
  bytesRead = fread ((void *) buffer, 1, sizeof (buffer) - 1, f);
  pclose (f);
  free (command);

  buffer[bytesRead] = 0;
  if (sizeA >= sizeB)
    diffBytes = atol (buffer) + (sizeA - sizeB);
  else
    diffBytes = atol (buffer) + (sizeB - sizeA);

  return (100 - (diffBytes * 100 / max (sizeA, sizeB)));
}

void
displayMatrix (int count, char **files)
{
  printf ("Comparing %d files. Mapping:\n\n", count);

  uint i, j, p;
  uint *diffMatrix;

  for (i = 0; i < count; i++)
    {
      printf (" %02d -> %s\n", (i + 1), files[i]);
    }

  printf ("\n");

  diffMatrix = (uint *) malloc (count * count * sizeof (uint));

  for (i = 0; i < count; i++)
    {
      for (j = i; j < count; j++)
	{
	  if (i == j)
	    {
	      diffMatrix[i * count + j] = 100;
	    }
	  else
	    {
	      diffMatrix[i * count + j] = getSimilarity (files[i], files[j]);
	    }
	}
    }

  printf ("    |");
  for (i = 0; i < count; i++)
    printf ("    %02d", i + 1);
  printf ("\n");

  printf ("----+");
  for (i = 0; i < count; i++)
    printf ("------");
  printf ("\n");


  for (i = 0; i < count; i++)
    {
      printf (" %02d |", i + 1);
      for (j = 0; j < i; j++)
	printf ("      ");
      for (j = i; j < count; j++)
	{
	  p = diffMatrix[i * count + j];
	  if (j == i)
	    pcolor (CL_WHITE);
	  else if (p >= 75)
	    pcolor (CL_GREEN);
	  else if (p >= 50)
	    pcolor (CL_YELLOW);
	  else
	    pcolor (CL_RED);

	  printf ("  %3d%%", diffMatrix[i * count + j]);

	  pcolor (CL_RESET);
	}

      printf ("\n");
    }

  printf ("\n");

  free (diffMatrix);
}

int
main (int argc, char **argv)
{
  if (argc < 2)
    {
      printf ("usage: %s FILE_1 FILE_2 ... FILE_n\n", argv[0]);
    }
  else
    {
      displayMatrix (argc - 1, &argv[1]);
    }

  return 0;
}
